import uuid
import datetime
import json
from pathlib import Path
import ifcopenshell
import ifcopenshell.api

from materialsdb.serialiser import XmlDeserialiser
from materialsdb import config
from materialsdb.classes import (
    Materials,
    Material,
    TLocalizedString,
    Webinfo,
    ISO639_1,
    Mimetype,
)

CATEGORIES = {
    "Others": {"hatch": "", "color": (255, 255, 255)},
    "Water_Proof": {"hatch": "", "color": (255, 255, 255)},
    "Vapour_Proof": {"hatch": "", "color": (0, 0, 0)},
    "Concrete": {"hatch": "", "color": (0, 255, 0)},
    "Wood_Timberproducts": {"hatch": "", "color": (91, 60, 17)},
    "Insulation": {"hatch": "", "color": (253, 108, 158)},
    "Masonry": {"hatch": "", "color": (253, 70, 38)},
    "Metal": {"hatch": "", "color": (119, 181, 254)},
    "Mortar": {"hatch": "", "color": (102, 0, 153)},
    "Plastics": {"hatch": "", "color": (96, 96, 96)},
    "Stone": {"hatch": "", "color": (0, 0, 255)},
    "Composite": {"hatch": "", "color": (112, 141, 35)},
    "Films": {"hatch": "", "color": (0, 0, 0)},
    "Render": {"hatch": "", "color": (0, 0, 0)},
    "Covering": {"hatch": "", "color": (0, 0, 0)},
    "Glas": {"hatch": "", "color": (27, 79, 8)},
    "Soil": {"hatch": "", "color": (142, 84, 52)},
}


def clean_psets(psets):
    new_dict = {}
    for pset_name, props in psets.items():
        pset_dict = {}
        for prop_name, definition in props.items():
            if definition["path"]:
                pset_dict[prop_name] = definition
        if pset_dict:
            new_dict[pset_name] = pset_dict
    return new_dict


PSETS = json.loads(Path(__file__).with_name("material_psets.json").read_text("utf-8"))
PSETS = clean_psets(PSETS)


def date_from_xml(days: float) -> datetime.datetime:
    """Convert days since 30.12.1899 which is materialsdb xml convention to datetime"""
    return datetime.datetime(1899, 12, 30) + datetime.timedelta(days=days)


def date_to_xml(date: datetime.datetime) -> float:
    """Convert datetime to days since 30.12.1899 which is materialsdb xml convention
    >>> datetime.timedelta(days=1).total_seconds()
    86400.0"""
    return (date - datetime.datetime(1899, 12, 30)).total_seconds() / 86400


def get_by_country(values, country):
    for value in values:
        if value.country == country:
            return value
    for value in values:
        if value.country is None:
            return value


def get_value(layer, definition, country=None):
    value = layer
    for attrib in definition["path"]:
        value = getattr(value, attrib)
        if isinstance(value, list):
            value = get_by_country(value, country)
            if not value:
                return None
    return value


class ProjectLibrary:
    def __init__(self, schema: str = "IFC4"):
        self.file = ifcopenshell.file(schema=schema)
        self.application = self.create_application()
        self.project_library = None
        self.lang = config.get_lang()
        self.country = config.get_country()
        self.owner_history = None

    def create_application(self):
        file = self.file

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcaddress.htm
        address = file.createIfcAddress(
            Purpose="OFFICE",
            Description="Anton Philipslaan 199\n5616TW Eindhoven\nThe Netherlands",
        )

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcorganization.htm
        organisation = file.createIfcOrganization(
            Name="AECGeeks",
            Description="""Software development and consultancy for the
            Architecture Engineering and Construction industry""",
            Addresses=[address],
        )

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcapplication.htm
        return file.createIfcApplication(
            ApplicationDeveloper=organisation,
            Version=ifcopenshell.version,
            ApplicationFullName="IfcOpenShell",
            ApplicationIdentifier="ifcopenshell",
        )

    def create_project_library(self, source: Materials, role: str = "MANUFACTURER"):
        file = self.file

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcroleenum.htm
        role = file.createIfcActorRole(role)

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcperson.htm
        person = file.createIfcPerson(
            Identification=str(uuid.uuid4()),
            FamilyName="Unknown",
            GivenName="Unknown",
            Roles=[role],
        )

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcorganization.htm
        organisation = file.createIfcOrganization(
            Identification=source.companyid, Name=source.company, Roles=[role]
        )

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcpersonandorganization.htm
        person_and_organisation = file.createIfcPersonAndOrganization(
            person, organisation, Roles=[role]
        )

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcownerhistory.htm
        owner_history = file.createIfcOwnerHistory(
            OwningUser=person_and_organisation,
            OwningApplication=self.application,
            CreationDate=int(date_from_xml(source.crd).timestamp()),
        )
        self.owner_history = owner_history

        # https://standards.buildingsmart.org/IFC/RELEASE/IFC4/ADD2_TC1/HTML/link/ifcprojectlibrary.htm
        file.createIfcProjectLibrary(
            GlobalId=ifcopenshell.guid.new(),
            OwnerHistory=owner_history,
            Name=source.company,
            Description=f"Material library converted from materialsdb xml for company {source.company}",
            ObjectType="MaterialLibrary",
            LongName=f"{source.company} version {source.ver}",
        )

    def create_materials(self, source: Materials):
        file = self.file
        context = file.createIfcRepresentationContext()
        for material in source.material:
            name = get_material_name(material, self.lang)
            description = get_material_description(material, self.lang)
            webinfo = get_material_webinfo(material, self.lang)
            category = material.information.group
            labels = material.information.labels
            color = material.information.color
            brushstyle = material.information.BrushStyle
            surface_style = self.get_surface_style(color, category)
            styled_item = file.createIfcStyledItem(Styles=[surface_style])
            # TODO: hatch_item = file.createIfcFillAreaStyleHatching()
            representation = file.createIfcStyledRepresentation(
                ContextOfItems=context,
                RepresentationIdentifier="Body",
                Items=[styled_item],
            )
            for layer in getattr(getattr(material, "layers", ()), "layer", ()):
                ifc_material = file.createIfcMaterial(name, description, category)
                for pset_name, props in PSETS.items():
                    properties = list()
                    for prop_name, definition in props.items():
                        primary_measure_type = definition["primary_measure_type"]
                        if not primary_measure_type:
                            continue
                        value = get_value(layer, definition, self.country)
                        if value:
                            unit_factor = definition.get("unit_factor", None) or 1
                            properties.append(
                                file.create_entity(
                                    "IfcPropertySingleValue",
                                    Name=prop_name,
                                    NominalValue=file.create_entity(
                                        primary_measure_type, value * unit_factor
                                    ),
                                )
                            )
                    if not properties:
                        continue
                    pset = file.create_entity(
                        "IfcMaterialProperties",
                        Name=pset_name,
                        Properties=properties,
                        Material=ifc_material,
                    )
                geometry = get_by_country(layer.geometry, self.country)
                if getattr(geometry, "thick", None):
                    element_name = name + f" | {geometry.thick}mm"
                    ifc_layer = file.create_entity(
                        "IfcMaterialLayer",
                        Material=ifc_material,
                        LayerThickness=geometry.thick / 1000,
                        Name=element_name,
                    )
                    assigned_material = file.create_entity(
                        "IfcMaterialLayerSet",
                        MaterialLayers=[ifc_layer],
                        LayerSetName=element_name,
                    )
                else:
                    element_name = name
                    assigned_material = ifc_material
                if material.information.wall:
                    wall = file.create_entity(
                        "IfcWallType",
                        GlobalId=ifcopenshell.guid.new(),
                        Name=element_name,
                    )
                    ifcopenshell.api.run(
                        "material.assign_material",
                        file,
                        product=wall,
                        material=assigned_material,
                    )
                if material.information.roof:
                    roof = file.create_entity(
                        "IfcRoofType",
                        GlobalId=ifcopenshell.guid.new(),
                        Name=element_name,
                    )
                    ifcopenshell.api.run(
                        "material.assign_material",
                        file,
                        product=roof,
                        material=assigned_material,
                    )
                if material.information.floor:
                    slab = file.create_entity(
                        "IfcRoofType",
                        GlobalId=ifcopenshell.guid.new(),
                        Name=element_name,
                    )
                    ifcopenshell.api.run(
                        "material.assign_material",
                        file,
                        product=slab,
                        material=assigned_material,
                    )
                if material.information.door:
                    door = file.create_entity(
                        "IfcDoorType",
                        GlobalId=ifcopenshell.guid.new(),
                        Name=element_name,
                    )
                    ifcopenshell.api.run(
                        "material.assign_material",
                        file,
                        product=door,
                        material=assigned_material,
                    )

    def get_surface_style(self, color, category):
        file = self.file
        if color:
            name = f"color {color}"
            style = file.createIfcSurfaceStyleShading(
                SurfaceColour=self.color_xml_to_ifc(color)
            )
        else:
            if not category:
                category = "Others"
            name = f"category {category}"
            for style in file.by_type("IfcSurfaceStyle"):
                if style.Name == name:
                    return style
            style = file.createIfcSurfaceStyleShading(
                SurfaceColour=self.file.createIfcColourRgb(
                    None, *CATEGORIES[category]["color"]
                ),
            )
        for surface_style in file.by_type("IfcSurfaceStyle"):
            if surface_style.Name == name:
                return surface_style
        return file.createIfcSurfaceStyle(Name=name, Side="BOTH", Styles=[style])

    def color_xml_to_ifc(self, color: int):
        """Color definition in xml is obscur. We assume that it is a decimal color.
        See: https://stackoverflow.com/a/2262152/4098083"""
        return self.file.createIfcColourRgb(
            Blue=color & 255, Green=(color >> 8) & 255, Red=(color >> 16) & 255
        )


def get_material_name(material: Material, lang: str) -> TLocalizedString:
    name = TLocalizedString("")
    for name in material.information.names.name:
        if name.lang == lang:
            return name
    return name


def get_material_description(material: Material, lang: str) -> TLocalizedString:
    description = TLocalizedString("")
    explanations = getattr(material.information, "explanations", None)
    for description in getattr(explanations, "explanation", ()):
        if description.lang == lang:
            return description
    return description


def get_material_webinfo(material: Material, lang: str) -> Webinfo:
    webinfo = Webinfo(Mimetype(""), "", "", lang=ISO639_1(lang))
    webinfos = getattr(material.information, "webinfos", None)
    for webinfo in getattr(webinfos, "webinfo", ()):
        if webinfo.lang == lang:
            return webinfo
    return webinfo


def create_project_library_from_xml(xml_path):
    library = ProjectLibrary()
    deserialiser = XmlDeserialiser()
    source = deserialiser.from_xml(str(xml_path))
    library.create_project_library(source)
    library.create_materials(source)
    return library.file


def main():
    file = create_project_library_from_xml("example_v103.xml")
    file.write("example_v103.xml")


if __name__ == "__main__":
    main()



import os
import pathlib
import json
from typing import Dict


def get_config_dir():
    config_dir = pathlib.Path(
        os.environ.get("APPDATA")
        or os.environ.get("XDG_CACHE_HOME")
        or pathlib.Path.home() / ".config"
    ).joinpath(
        "materialsdb",
    )
    pathlib.Path(config_dir).mkdir(parents=True, exist_ok=True)
    return config_dir


def get_base_config_path() -> pathlib.Path:
    return get_config_dir() / "config.json"


def get_base_config() -> Dict[str, str]:
    base_config_path = get_base_config_path()
    if base_config_path.exists():
        return json.loads(get_base_config_path().read_text("utf-8"))
    return {"lang": "fr", "country": "CH"}


def get_param(param: str) -> str:
    return get_base_config()[param]


def set_param(param: str, value: str) -> None:
    config = get_base_config()
    config[param] = value
    get_base_config_path().write_text(json.dumps(config), encoding="utf-8")


def get_lang() -> str:
    return get_param("lang")


def set_lang(lang: str):
    set_param("lang", lang)


def get_country() -> str:
    return get_param("country")


def set_country(country: str):
    set_param("country", country)


def main():
    lang = get_lang()
    set_lang(lang)
    country = get_country()
    set_country(country)


if __name__ == "__main__":
    main()



import datetime
from materialsdb.classes import TDateTime


def date_from_xml(days: float) -> datetime.datetime:
    """Convert days since 30.12.1899 which is materialsdb xml convention to datetime"""
    return datetime.datetime(1899, 12, 30) + datetime.timedelta(days=days)


def date_to_xml(date: datetime.datetime) -> float:
    """Convert datetime to days since 30.12.1899 which is materialsdb xml convention
    >>> datetime.timedelta(days=1).total_seconds()
    86400.0"""
    return (
        date - datetime.datetime(1899, 12, 30, tzinfo=datetime.timezone.utc)
    ).total_seconds() / 86400


def new_tdatetime():
    return TDateTime(date_to_xml(datetime.datetime.now(datetime.timezone.utc)))



import os
import pathlib
import urllib.request
from typing import Optional
from lxml import etree

MATERIALSDBINDEXURL = "http://www.materialsdb.org/download/ProducerIndex.xml"


def get_cache_folder():
    cache_dir = pathlib.Path(
        os.environ.get("APPDATA")
        or os.environ.get("XDG_CACHE_HOME")
        or pathlib.Path.home() / ".cache"
    ).joinpath(
        "materialsdb",
    )
    pathlib.Path(cache_dir).mkdir(parents=True, exist_ok=True)
    return cache_dir


def get_cached_index_path() -> pathlib.Path:
    return get_cache_folder() / "ProducerIndex.xml"


def parse_cached_index() -> etree._ElementTree:
    path = get_cached_index_path()
    if path.exists():
        return etree.parse(str(path))
    root = etree.Element("root")
    return etree.ElementTree(root)


def get_by_id(root: etree._Element, id: str) -> Optional[etree._Element]:
    for company in root:
        if company.get("id") == id:
            return company
    return None


def require_update(cached_company, company) -> bool:
    if cached_company is None:
        return True
    for attrib in ["LastKnownDate", "KnownVersion"]:
        if company.get(attrib) > cached_company.get(attrib):
            return True
    return False


def get_producers_dir() -> pathlib.Path:
    producer_path = get_cache_folder().joinpath("Producers")
    pathlib.Path(producer_path).mkdir(parents=True, exist_ok=True)
    return producer_path


def update_producers_data():
    cached_index = parse_cached_index()
    cached_root = cached_index.getroot()
    new_index = etree.parse(MATERIALSDBINDEXURL)
    new_root = new_index.getroot()
    producers_dir = get_producers_dir()
    has_index_update = False
    for company in new_root:
        cached_company = get_by_id(cached_root, company.get("id"))
        if (
            not require_update(cached_company, company)
            and (producers_dir / pathlib.Path(company.get("href")).name).exists()
        ):
            continue
        if cached_company:
            (producers_dir / pathlib.Path(cached_company.get("href")).name).unlink(True)
        has_index_update = True
        producer_path = producers_dir / pathlib.Path(company.get("href")).name
        urllib.request.urlretrieve(company.get("href"), producer_path)
    if has_index_update:
        new_index.write(str(get_cached_index_path()))


def producers():
    for producer in get_producers_dir().iterdir():
        if producer.suffix.lower() == ".xml":
            yield producer


def main():
    update_producers_data()


if __name__ == "__main__":
    main()



# coding: UTF-8
""" This file contain classes corresponding to materialsdb*.xsd elements.
This file is automatically generated. Do not modify it directly.
If there is an issue or an enhancement to make then do it in dataclasses_generator.py

© All rights reserved.
ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE, Switzerland, Laboratory CNPA, 2019-2020

See the LICENSE.md file for more details.

Author : Cyril Waechter
"""
from dataclasses import dataclass
from typing import Tuple, Optional, List


class TLCADB(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('dbKBOB', 'dbOkobauDat', 'dbLux')
    xml_name: str = "TLCADB"


class TFireComb(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('3', '4', '5', '5_200C', '6q', '6')
    xml_name: str = "TFireComb"


class TFireSmoke(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('1', '2', '3')
    xml_name: str = "TFireSmoke"


class TFireResis(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('A1', 'A2', 'B', 'C', 'D', 'E', 'F')
    xml_name: str = "TFireResis"


class TFireReact(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('RF1', 'RF2', 'RF2cr', 'RF3', 'RF3cr', 'RF4cr')
    xml_name: str = "TFireReact"


class TFireDin4102(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('A1', 'A2', 'B1', 'B2', 'B3')
    xml_name: str = "TFireDin4102"


class T2Lowercase(str):
    xs_type: str = "simpleType"
    xml_pattern:str = "[a-z][a-z]"
    xml_name: str = "T2Lowercase"


class Groupkind(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('Others', 'Water_Proof', 'Vapour_Proof', 'Concrete', 'Wood_Timberproducts', 'Insulation', 'Masonry', 'Metal', 'Mortar', 'Plastics', 'Stone', 'Composite', 'Films', 'Render', 'Covering', 'Glas', 'Soil', 'air')
    xml_name: str = "groupkind"


class Materialtype(str):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('simple', 'btk', 'construction')
    xml_name: str = "materialtype"


class T2Uppercase(str):
    xs_type: str = "simpleType"
    xml_pattern:str = "[A-Z][A-Z]"
    xml_name: str = "T2Uppercase"


class TDateTime(float):
    xs_type: str = "simpleType"
    xml_name: str = "TDateTime"


class Guid(str):
    xs_type: str = "simpleType"
    xml_pattern:str = "[0-9a-fA-F]{8}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{4}-[0-9a-fA-F]{12}"
    xml_name: str = "guid"


class Brackguid(str):
    xs_type: str = "simpleType"
    xml_pattern:str = "\{[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}\}"
    xml_name: str = "Brackguid"


class Boolean(int):
    xs_type: str = "simpleType"
    xml_pattern:str = "[0-1]"
    xml_name: str = "boolean"


class Mimetype(str):
    xs_type: str = "simpleType"
    xml_name: str = "mimetype"


class ISO639_1(T2Lowercase):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('aa', 'ab', 'af', 'am', 'ar', 'as', 'ay', 'az', 'ba', 'be', 'bg', 'bh', 'bi', 'bn', 'bo', 'br', 'ca', 'co', 'cs', 'cy', 'da', 'de', 'dz', 'el', 'en', 'eo', 'es', 'et', 'eu', 'fa', 'fi', 'fj', 'fo', 'fr', 'fy', 'ga', 'gd', 'gl', 'gn', 'gu', 'ha', 'he', 'hi', 'hr', 'hu', 'hy', 'ia', 'ie', 'ik', 'id', 'is', 'it', 'iu', 'ja', 'jv', 'ka', 'kk', 'kl', 'km', 'kn', 'ko', 'ks', 'ku', 'ky', 'la', 'ln', 'lo', 'lt', 'lv', 'mg', 'mi', 'mk', 'ml', 'mn', 'mo', 'mr', 'ms', 'mt', 'my', 'na', 'ne', 'nl', 'no', 'oc', 'om', 'or', 'pa', 'pl', 'ps', 'pt', 'qu', 'rm', 'rn', 'ro', 'ru', 'rw', 'sa', 'sd', 'sg', 'sh', 'si', 'sk', 'sl', 'sm', 'sn', 'so', 'sq', 'sr', 'ss', 'st', 'su', 'sv', 'sw', 'ta', 'te', 'tg', 'th', 'ti', 'tk', 'tl', 'tn', 'to', 'tr', 'ts', 'tt', 'tw', 'ug', 'uk', 'ur', 'uz', 'vi', 'vo', 'wo', 'xh', 'yi', 'yo', 'za', 'zh', 'zu')
    xml_name: str = "ISO639_1"


class ISO_3166_1_alpha_2(T2Uppercase):
    xs_type: str = "simpleType"
    xml_enum: Tuple[str, ...] = ('AD', 'AE', 'AF', 'AG', 'AI', 'AL', 'AM', 'AN', 'AO', 'AQ', 'AR', 'AS', 'AT', 'AU', 'AW', 'AZ', 'BA', 'BB', 'BD', 'BE', 'BF', 'BG', 'BH', 'BI', 'BJ', 'BM', 'BN', 'BO', 'BR', 'BS', 'BT', 'BV', 'BW', 'BY', 'BZ', 'CA', 'CC', 'CD', 'CF', 'CG', 'CH', 'CI', 'CK', 'CL', 'CM', 'CN', 'CO', 'CR', 'CS', 'CU', 'CV', 'CX', 'CY', 'CZ', 'DE', 'DJ', 'DK', 'DM', 'DO', 'DZ', 'EC', 'EE', 'EG', 'EH', 'ER', 'ES', 'ET', 'FI', 'FJ', 'FK', 'FM', 'FO', 'FR', 'GA', 'GB', 'GD', 'GE', 'GF', 'GH', 'GI', 'GL', 'GM', 'GN', 'GP', 'GQ', 'GR', 'GS', 'GT', 'GU', 'GW', 'GY', 'HK', 'HM', 'HN', 'HR', 'HT', 'HU', 'ID', 'IE', 'IL', 'IN', 'IO', 'IQ', 'IR', 'IS', 'IT', 'JM', 'JO', 'JP', 'KE', 'KG', 'KH', 'KI', 'KM', 'KN', 'KP', 'KR', 'KW', 'KY', 'KZ', 'LA', 'LB', 'LC', 'LI', 'LK', 'LR', 'LS', 'LT', 'LU', 'LV', 'LY', 'MA', 'MC', 'MD', 'MG', 'MH', 'MK', 'ML', 'MM', 'MN', 'MO', 'MP', 'MQ', 'MR', 'MS', 'MT', 'MU', 'MV', 'MW', 'MX', 'MY', 'MZ', 'NA', 'NC', 'NE', 'NF', 'NG', 'NI', 'NL', 'NO', 'NP', 'NR', 'NU', 'NZ', 'OM', 'PA', 'PE', 'PF', 'PG', 'PH', 'PK', 'PL', 'PM', 'PN', 'PR', 'PS', 'PT', 'PW', 'PY', 'QA', 'RE', 'RO', 'RU', 'RW', 'SA', 'SB', 'SC', 'SD', 'SE', 'SG', 'SH', 'SI', 'SJ', 'SK', 'SL', 'SM', 'SN', 'SO', 'SR', 'ST', 'SV', 'SY', 'SZ', 'TC', 'TD', 'TF', 'TG', 'TH', 'TJ', 'TK', 'TL', 'TM', 'TN', 'TO', 'TR', 'TT', 'TV', 'TW', 'TZ', 'UA', 'UG', 'UM', 'US', 'UY', 'UZ', 'VA', 'VC', 'VE', 'VG', 'VI', 'VN', 'VU', 'WF', 'WS', 'YE', 'YT', 'ZA', 'ZM', 'ZW')
    xml_name: str = "ISO_3166_1_alpha_2"


class THexHash(str):
    ver: int
    xs_type: str = "complexType"
    xml_attributes: Tuple[str, ...] = ('ver',)
    xml_name: str = "THexHash"
    def __new__(cls, object, *args, **kwargs):
        obj = str.__new__(cls, object)
        for arg_name, arg_value in zip(['ver'], args):
            arg_name = arg_value
        for key, value in kwargs.items():
            if key == "object":
                continue
            setattr(obj, key, value)
        return obj


class TLocalizedString(str):
    xs_type: str = "complexType"
    lang: Optional[ISO639_1] = None
    xml_attributes: Tuple[str, ...] = ('lang',)
    xml_name: str = "TLocalizedString"
    def __new__(cls, object, *args, **kwargs):
        obj = str.__new__(cls, object)
        for arg_name, arg_value in zip(['lang'], args):
            arg_name = arg_value
        for key, value in kwargs.items():
            if key == "object":
                continue
            setattr(obj, key, value)
        return obj


class TCountryLocalizedString(TLocalizedString):
    xs_type: str = "complexType"
    lang: Optional[ISO639_1] = None
    country: Optional[ISO_3166_1_alpha_2] = None
    xml_attributes: Tuple[str, ...] = ('lang', 'country')
    xml_name: str = "TCountryLocalizedString"
    def __new__(cls, object, *args, **kwargs):
        obj = TLocalizedString.__new__(cls, object)
        for arg_name, arg_value in zip(['lang', 'country'], args):
            arg_name = arg_value
        for key, value in kwargs.items():
            if key == "object":
                continue
            setattr(obj, key, value)
        return obj


class TNamedString(str):
    name: str
    xs_type: str = "complexType"
    xml_attributes: Tuple[str, ...] = ('name',)
    xml_name: str = "TNamedString"
    def __new__(cls, object, *args, **kwargs):
        obj = str.__new__(cls, object)
        for arg_name, arg_value in zip(['name'], args):
            arg_name = arg_value
        for key, value in kwargs.items():
            if key == "object":
                continue
            setattr(obj, key, value)
        return obj


@dataclass
class Names:
    name: List[TCountryLocalizedString]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "names"
    xml_elements: Tuple[str, ...] = ('name',)


@dataclass
class Labels:
    label: List[TNamedString]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "labels"
    xml_elements: Tuple[str, ...] = ('label',)


@dataclass
class Country:
    sellingfrom: TDateTime
    xs_type: str = "element"
    name: Optional[ISO_3166_1_alpha_2] = None
    sellinguntil: Optional[TDateTime] = None
    xml_attributes: Tuple[str, ...] = ('name', 'sellingfrom', 'sellinguntil')
    xml_name = "country"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Countries:
    country: List[Country]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "countries"
    xml_elements: Tuple[str, ...] = ('country',)


@dataclass
class Explanations:
    explanation: List[TLocalizedString]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "explanations"
    xml_elements: Tuple[str, ...] = ('explanation',)


@dataclass
class Webinfo:
    href: str
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    lang: Optional[ISO639_1] = None
    mime: Optional[Mimetype] = None
    size: Optional[float] = None
    title: Optional[str] = None
    xml_attributes: Tuple[str, ...] = ('country', 'lang', 'mime', 'size', 'title', 'href')
    xml_name = "webinfo"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Webinfos:
    webinfo: List[Webinfo]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "webinfos"
    xml_elements: Tuple[str, ...] = ('webinfo',)


@dataclass
class Companyref:
    xs_type: str = "element"
    reftype: Optional[str] = None
    xml_attributes: Tuple[str, ...] = ('reftype',)
    xml_name = "companyref"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Companyrefs:
    companyref: List[Companyref]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "companyrefs"
    xml_elements: Tuple[str, ...] = ('companyref',)


@dataclass
class Information:
    names: Names
    xs_type: str = "element"
    group: Optional[Groupkind] = Groupkind("Others")
    wall: Optional[Boolean] = Boolean("0")
    roof: Optional[Boolean] = Boolean("0")
    floor: Optional[Boolean] = Boolean("0")
    door: Optional[Boolean] = Boolean("0")
    color: Optional[int] = None
    BrushStyle: Optional[int] = None
    xml_attributes: Tuple[str, ...] = ('group', 'wall', 'roof', 'floor', 'door', 'color', 'BrushStyle')
    labels: Optional[Labels] = None
    countries: Optional[Countries] = None
    explanations: Optional[Explanations] = None
    webinfos: Optional[Webinfos] = None
    companyrefs: Optional[Companyrefs] = None
    xml_name = "information"
    xml_elements: Tuple[str, ...] = ('names', 'labels', 'countries', 'explanations', 'webinfos', 'companyrefs')


@dataclass
class Geometry:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    thick: Optional[float] = None
    length: Optional[str] = None
    width: Optional[str] = None
    design: Optional[str] = None
    xml_attributes: Tuple[str, ...] = ('country', 'thick', 'length', 'width', 'design')
    xml_name = "geometry"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Thermal:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    lambda_value: Optional[float] = None
    lambda_value_dry: Optional[float] = None
    mu_min: Optional[float] = None
    mu_max: Optional[float] = None
    therm_capa: Optional[float] = None
    W80_coef: Optional[float] = None
    Wf_coef: Optional[float] = None
    A_coef: Optional[float] = None
    B_coef: Optional[float] = None
    Allwd_Moisture: Optional[float] = None
    xml_attributes: Tuple[str, ...] = ('country', 'lambda_value', 'lambda_value_dry', 'mu_min', 'mu_max', 'therm_capa', 'W80_coef', 'Wf_coef', 'A_coef', 'B_coef', 'Allwd_Moisture')
    xml_name = "thermal"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Physical:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    MinElas: Optional[float] = None
    MaxElas: Optional[float] = None
    ResCompMin: Optional[float] = None
    ResTractMin: Optional[float] = None
    density: Optional[float] = None
    T: Optional[str] = None
    CP: Optional[str] = None
    WS: Optional[str] = None
    Porosity: Optional[float] = None
    DS_TH: Optional[str] = None
    xml_attributes: Tuple[str, ...] = ('country', 'MinElas', 'MaxElas', 'ResCompMin', 'ResTractMin', 'density', 'T', 'CP', 'WS', 'Porosity', 'DS_TH')
    xml_name = "physical"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Security:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    FireComb: Optional[TFireComb] = None
    FireSmoke: Optional[TFireSmoke] = None
    FireResis: Optional[TFireResis] = None
    FireReact: Optional[TFireReact] = None
    Baustoffklasse: Optional[TFireDin4102] = None
    FireClass: Optional[str] = None
    xml_attributes: Tuple[str, ...] = ('country', 'FireComb', 'FireSmoke', 'FireResis', 'FireReact', 'Baustoffklasse', 'FireClass')
    xml_name = "security"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Acoustic:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    n125: Optional[float] = None
    n250: Optional[float] = None
    n500: Optional[float] = None
    n1000: Optional[float] = None
    n2000: Optional[float] = None
    n4000: Optional[float] = None
    free_space: Optional[float] = None
    SD: Optional[float] = None
    air_flow_resistivity: Optional[float] = None
    xml_attributes: Tuple[str, ...] = ('country', 'n125', 'n250', 'n500', 'n1000', 'n2000', 'n4000', 'free_space', 'SD', 'air_flow_resistivity')
    xml_name = "acoustic"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Other:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    Rademissivity: Optional[float] = None
    Radabsorp: Optional[float] = None
    usetemplimit: Optional[float] = None
    ProductCode: Optional[str] = None
    materialtype: Optional[str] = None
    TargetPrice: Optional[float] = None
    xml_attributes: Tuple[str, ...] = ('country', 'Rademissivity', 'Radabsorp', 'usetemplimit', 'ProductCode', 'materialtype', 'TargetPrice')
    xml_name = "other"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Lcia:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    UBP: Optional[float] = None
    NRE: Optional[float] = None
    GWP: Optional[float] = None
    CED: Optional[float] = None
    IPRIM: Optional[float] = None
    IENV: Optional[float] = None
    IECO: Optional[float] = None
    IECO12: Optional[float] = None
    KBOB_eq: Optional[Brackguid] = None
    xml_attributes: Tuple[str, ...] = ('country', 'UBP', 'NRE', 'GWP', 'CED', 'IPRIM', 'IENV', 'IECO', 'IECO12', 'KBOB_eq')
    xml_name = "lcia"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Lcaversion:
    id: str
    code: Brackguid
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ('id', 'code')
    xml_name = "lcaversion"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Lca:
    database: TLCADB
    xs_type: str = "element"
    luxInsulation: Optional[Guid] = None
    xml_attributes: Tuple[str, ...] = ('database', 'luxInsulation')
    lcaversion: Optional[List[Lcaversion]] = None
    xml_name = "lca"
    xml_elements: Tuple[str, ...] = ('lcaversion',)


@dataclass
class Emission:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    TVOC: Optional[float] = None
    Formaldehide: Optional[float] = None
    Ammonia: Optional[float] = None
    Carginogenic: Optional[float] = None
    xml_attributes: Tuple[str, ...] = ('country', 'TVOC', 'Formaldehide', 'Ammonia', 'Carginogenic')
    xml_name = "emission"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Layer:
    id: Guid
    xs_type: str = "element"
    aliases: Optional[str] = None
    displayorder: Optional[int] = None
    xml_attributes: Tuple[str, ...] = ('id', 'aliases', 'displayorder')
    geometry: Optional[List[Geometry]] = None
    thermal: Optional[List[Thermal]] = None
    physical: Optional[List[Physical]] = None
    security: Optional[List[Security]] = None
    acoustic: Optional[List[Acoustic]] = None
    other: Optional[List[Other]] = None
    lcia: Optional[List[Lcia]] = None
    lca: Optional[List[Lca]] = None
    emission: Optional[List[Emission]] = None
    xml_name = "layer"
    xml_elements: Tuple[str, ...] = ('geometry', 'thermal', 'physical', 'security', 'acoustic', 'other', 'lcia', 'lca', 'emission')


@dataclass
class Layers:
    layer: List[Layer]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "layers"
    xml_elements: Tuple[str, ...] = ('layer',)


@dataclass
class Vgeometry:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    thick: Optional[float] = None
    density: Optional[float] = None
    xml_attributes: Tuple[str, ...] = ('country', 'thick', 'density')
    xml_name = "vgeometry"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Vthermal:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    U_value_without: Optional[float] = None
    U_value_with: Optional[float] = None
    ETA1: Optional[float] = None
    ETA2: Optional[float] = None
    U24: Optional[float] = None
    CM1: Optional[float] = None
    CM2: Optional[float] = None
    metalcladding: Optional[Boolean] = Boolean("0")
    xml_attributes: Tuple[str, ...] = ('country', 'U_value_without', 'U_value_with', 'ETA1', 'ETA2', 'U24', 'CM1', 'CM2', 'metalcladding')
    xml_name = "vthermal"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Vacoustic:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    lang: Optional[ISO639_1] = None
    ExplainAcoustic: Optional[str] = None
    RwCalc: Optional[float] = None
    RwPrimeCalc: Optional[float] = None
    RwMesured: Optional[float] = None
    RwPrimeMesured: Optional[float] = None
    xml_attributes: Tuple[str, ...] = ('country', 'lang', 'ExplainAcoustic', 'RwCalc', 'RwPrimeCalc', 'RwMesured', 'RwPrimeMesured')
    xml_name = "vacoustic"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Vother:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    ProductCode: Optional[str] = None
    Rademissivity: Optional[float] = None
    Radabsorp: Optional[float] = None
    xml_attributes: Tuple[str, ...] = ('country', 'ProductCode', 'Rademissivity', 'Radabsorp')
    xml_name = "vother"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Vlcia:
    xs_type: str = "element"
    country: Optional[ISO_3166_1_alpha_2] = None
    UBP: Optional[float] = None
    NRE: Optional[float] = None
    GWP: Optional[float] = None
    CED: Optional[float] = None
    IPRIM: Optional[float] = None
    IENV: Optional[float] = None
    IECO: Optional[float] = None
    IECO12: Optional[float] = None
    KBOB_eq: Optional[Brackguid] = None
    xml_attributes: Tuple[str, ...] = ('country', 'UBP', 'NRE', 'GWP', 'CED', 'IPRIM', 'IENV', 'IECO', 'IECO12', 'KBOB_eq')
    xml_name = "vlcia"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Vlca:
    database: TLCADB
    xs_type: str = "element"
    luxInsulation: Optional[Guid] = None
    xml_attributes: Tuple[str, ...] = ('database', 'luxInsulation')
    lcaversion: Optional[List[Lcaversion]] = None
    xml_name = "vlca"
    xml_elements: Tuple[str, ...] = ('lcaversion',)


@dataclass
class Variation:
    id: Guid
    vgeometry: List[Vgeometry]
    vthermal: List[Vthermal]
    xs_type: str = "element"
    displayorder: Optional[int] = None
    xml_attributes: Tuple[str, ...] = ('id', 'displayorder')
    vacoustic: Optional[List[Vacoustic]] = None
    vother: Optional[List[Vother]] = None
    vlcia: Optional[List[Vlcia]] = None
    vlca: Optional[List[Vlca]] = None
    xml_name = "variation"
    xml_elements: Tuple[str, ...] = ('vgeometry', 'vthermal', 'vacoustic', 'vother', 'vlcia', 'vlca')


@dataclass
class Variations:
    variation: List[Variation]
    xs_type: str = "element"
    xml_attributes: Tuple[str, ...] = ()
    xml_name = "variations"
    xml_elements: Tuple[str, ...] = ('variation',)


@dataclass
class Construction:
    xs_type: str = "element"
    source: Optional[str] = None
    consref: Optional[str] = None
    xml_attributes: Tuple[str, ...] = ('source', 'consref')
    xml_name = "construction"
    xml_elements: Tuple[str, ...] = ()


@dataclass
class Material:
    id: Guid
    type: Materialtype
    information: Information
    xs_type: str = "element"
    readonly: Optional[Boolean] = Boolean("1")
    xml_attributes: Tuple[str, ...] = ('id', 'type', 'readonly')
    layers: Optional[Layers] = None
    variations: Optional[Variations] = None
    construction: Optional[Construction] = None
    xml_name = "material"
    xml_elements: Tuple[str, ...] = ('information', 'layers', 'variations', 'construction')


@dataclass
class Materials:
    company: str
    companyid: Guid
    ver: int
    crd: TDateTime
    verXML: int
    material: List[Material]
    sig: THexHash
    publickey: THexHash
    xs_type: str = "element"
    SchemaLocation: Optional[str] = None
    xml_attributes: Tuple[str, ...] = ('company', 'companyid', 'ver', 'crd', 'verXML', 'SchemaLocation')
    xml_name = "materials"
    xml_elements: Tuple[str, ...] = ('material', 'sig', 'publickey')




# coding: UTF-8
"""This module serialise/deserialise materials from different schema/format.

© All rights reserved.
ECOLE POLYTECHNIQUE FEDERALE DE LAUSANNE, Switzerland, Laboratory CNPA, 2019-2020

See the LICENSE.md file for more details.

Author : Cyril Waechter
"""
import re
from pathlib import Path
import typing
from typing import Protocol, Tuple, Dict, Type, Optional, Any, Union

from lxml import objectify, etree

from materialsdb import classes


def get_xml_schema() -> str:
    return str(Path(__file__).parent / "schema/materialsdb103.xsd")


def get_element_name(element: objectify.ObjectifiedElement) -> str:
    m = re.search("{.*}(.*)", element.tag)
    return m.group(1) if m else element.tag


def create_element_maker():
    nsmap = {
        None: "http://www.materialsdb.org",
        "xsi": "http://www.w3.org/2001/XMLSchema-instance",
        "SchemaLocation": "http://www.materialsdb.org/schemas/materialsdb103.xsd",
    }
    return objectify.ElementMaker(
        namespace="http://www.materialsdb.org", nsmap=nsmap, annotate=False
    )


def get_valid_root(tree: objectify.ObjectifiedElement) -> str:
    root = tree.getroot()
    for attr in ("sig", "publickey"):
        if not hasattr(root, attr):
            oem = create_element_maker()
            print(
                f"{Path(root.base).name}: Element '{attr}' is missing from '{root.tag}'"
            )
            root.append(oem(attr, "Missing '{attr}'", {"ver": 0}))
    return root


class XmlDeserialiser:
    def __init__(self):
        self.schema = etree.XMLSchema(file=get_xml_schema())
        self.parser = objectify.makeparser(schema=self.schema)

    def from_xml(self, xml_path: str, assert_schema: bool = False):
        if assert_schema:
            tree = objectify.parse(xml_path, self.parser)
        else:
            tree = objectify.parse(xml_path)
        return self.from_element(get_valid_root(tree))

    def from_element(self, element=None, base_class=None):
        element_name = get_element_name(element)
        element_class = base_class or getattr(classes, self.cls_name(element_name))
        kwargs: Dict[str, Any] = {}
        if element_class.xs_type != "element":
            kwargs["object"] = element.text or ""
        type_hints = typing.get_type_hints(element_class)
        for attrib in getattr(element_class, "xml_attributes", ()):
            value = element.get(attrib)
            if value is None:
                continue
            base_class = self.strip_optional(type_hints[attrib])

            kwargs[attrib] = base_class(value)
        for child_name in getattr(element_class, "xml_elements", ()):
            type_hint = type_hints[child_name]
            base_class = self.strip_optional(type_hint)
            if typing.get_origin(base_class) is list:
                value = []
                inner_type = typing.get_args(base_class)[0]
                for child in getattr(element, child_name, ()):
                    try:
                        value.append(self.from_element(child, inner_type))
                    except TypeError as err:
                        print(
                            f"{Path(child.base).name}: Element '{child_name}' line {child.sourceline} seems invalid:\n\t{err}"
                        )
                kwargs[child_name] = value
            else:
                child = getattr(element, child_name, None)
                if child is not None:
                    kwargs[child_name] = self.from_element(child, base_class)

        return element_class(**kwargs)

    def strip_optional(self, type_hint):
        if self.is_optional(type_hint):
            return typing.get_args(type_hint)[0]
        return type_hint

    def is_optional(self, type_hint):
        return typing.get_origin(type_hint) is typing.Union and typing.get_args(
            type_hint
        )[1] is type(None)

    def cls_name(self, name: str) -> str:
        return name if name[0].isupper() else name.title()


class XmlSerialiser:
    def __init__(self):
        self.schema = etree.XMLSchema(file=get_xml_schema())
        self.parser = objectify.makeparser(schema=self.schema)
        self.oem = create_element_maker()

    def to_xml(self, source: classes.Materials, xml_path: str) -> None:
        xml_root = self.serialise(source)
        objectify.deannotate(xml_root, xsi_nil=True)
        xml_str = etree.tostring(
            xml_root,
            encoding="UTF-8",
            xml_declaration=True,
            pretty_print=True,
            standalone=False,
        )
        with open(xml_path, "wb") as file:
            file.write(xml_str)

    def serialise(self, element, name: str = None):
        # Process base element with attribute and text value
        kwargs = {}
        for attrib in getattr(element, "xml_attributes", ()):
            value = getattr(element, attrib)
            if value is None:
                continue
            kwargs[attrib] = str(value)
        text = element if element.xs_type != "element" else None
        name = name or element.xml_name
        xml_element = self.oem(name, text, **kwargs)
        # Process child elements and their childs recursively
        for child_name in getattr(element, "xml_elements", ()):
            childs = getattr(element, child_name, None)
            if childs is None:
                continue
            if not isinstance(childs, list):
                childs = (childs,)
            for child in childs:
                xml_element.append(self.serialise(child, child_name))

        return xml_element


def main():
    xml_path = "example_v103.xml"
    deserialiser = XmlDeserialiser()
    sources = []
    sources.append(deserialiser.from_xml(xml_path))
    serialiser = XmlSerialiser()
    serialiser.to_xml(sources[0], xml_path="test.xml")


if __name__ == "__main__":
    main()
